import openai
import json
import time
import os
import subprocess
import shutil
import ocrmypdf
import glob

from termcolor import colored, cprint
from rich.console import Console
from rich import print as rprint
from rich.panel import Panel
from rich.text import Text
from rich.table import Table


class bcolors:
    HEADER = "\033[95m"
    OKBLUE = "\033[94m"
    OKCYAN = "\033[96m"
    OKGREEN = "\033[92m"
    WHITE = "\33[97m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"


class Chat:
    character = {
        "name": "Default Assistant",  # default name
        "system": "You are a helpful assistant.",
        "temperature": 0,
        "voice": "Evan (Enhanced)",
    }

    character_presets = {
        # 'hal9000': 'characters/hal9000.json'
    }

    user_home = os.path.expanduser("~")
    config_root_path = None
    config_character_path = None
    history_path = None

    def __init__(self):
        self.config_root_path = os.path.join(self.user_home, ".config", "termq")
        self.config_character_path = os.path.join(self.config_root_path, "characters")
        self.history_path = os.path.join(self.config_root_path, "history")

        self.generateConfig()
        self.updateCharacterPresets()

        # load the character presets
        self.character_presets = {}
        json_files = glob.glob(os.path.join(self.config_character_path, "*.json"))
        for file in json_files:
            preset_name = os.path.splitext(os.path.basename(file))[0]
            self.character_presets[preset_name] = file

    def generateConfig(self):
        """Create necessary directory structure for termq."""
        os.makedirs(self.config_root_path, exist_ok=True)
        os.makedirs(self.config_character_path, exist_ok=True)
        os.makedirs(self.history_path, exist_ok=True)

    def copy_files(self, source_dir, destination_dir):
        # Get the list of files in the source directory
        files = os.listdir(source_dir)

        # Iterate over each file and copy it to the destination directory
        for file_name in files:
            source_file = os.path.join(source_dir, file_name)
            destination_file = os.path.join(destination_dir, file_name)
            shutil.copy2(source_file, destination_file)

    def updateCharacterPresets(self):
        # Get the path to the package's characters directory
        package_dir = os.path.dirname(os.path.abspath(__file__))
        characters_source = os.path.join(package_dir, "characters")

        # Only update if source characters directory exists
        if not os.path.exists(characters_source):
            return

        self.copy_files(characters_source, self.config_character_path)

    def getHistoryFilePath(self, chat_datetime):
        return os.path.join(self.history_path, f"chat_history-{chat_datetime}.json")

    def loadCharacters(self, character_file):
        if character_file in self.character_presets:
            character_file = self.character_presets[character_file]

        # check if the character profile exists
        if not os.path.isfile(character_file) or not os.path.isfile(
            character_file.replace(".json", ".txt")
        ):
            error_msg = [
                f"The preset file path you specified does not exist: `{character_file}`"
            ]
            rprint(
                Panel(
                    "\n".join(error_msg),
                    title="TermQ",
                    border_style="red bold",
                    style="red",
                )
            )
            exit(0)

        # load the character profile
        with open(character_file) as f:
            character_profile = json.load(f)
            self.character["name"] = character_profile["name"]
            self.character["temperature"] = character_profile["temperature"]
            self.character["voice"] = character_profile["voice"]

        # load the character's prompt
        with open(character_file.replace(".json", ".txt")) as f:
            self.character["system"] = "".join(f.read())

    def getInitialMessage(self):
        return [{"role": "system", "content": self.character["system"]}]

    def welcome(self, engine_type):
        # Checking OPENAI_API_KEY
        openai.api_key = os.getenv("OPENAI_API_KEY")
        if openai.api_key is None:
            error_msg = ["OPENAI_API_KEY is not set as a system environment variable"]
            rprint(
                Panel(
                    "\n".join(error_msg),
                    title="TermQ",
                    border_style="red bold",
                    style="red",
                    # subtitle="Thank you"
                )
            )
            exit(0)

        welcome_msg = [
            # f"[SYSTEM] Using `{engine_type}`",
            f"[SYSTEM] {self.character['name']} is ready to chat.",
            f"[SYSTEM] Chat dialogue will be saved to `{self.history_path}`",
            "[SYSTEM] <Press Ctrl+C to exit>",
        ]
        rprint(
            Panel(
                "\n".join(welcome_msg),
                title="TermQ",
                border_style="white bold",
                style="white",
                # subtitle="Thank you"
            )
        )

    def showGoodbyeMessage(self):
        goodbye_msg = [
            "Bye",
        ]
        print()
        rprint(
            Panel(
                "\n".join(goodbye_msg),
                title="TermQ",
                border_style="red bold",
                style="red",
            )
        )

    def say(self, message, voice="Samantha"):
        # say -v \? | more
        stdout, stderr = self.exec("say " + '-v "' + voice + '" "' + message + '"')

    def exec(self, command):
        # Execute the command and capture the output
        process = subprocess.Popen(
            command,
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )

        stdout = ""
        stderr = ""

        try:
            # Continuously read the output
            while True:
                # Read the output
                output = process.stdout.readline()
                errorp = process.stderr.readline()

                # Check if the output is empty and the process has finished
                if process.poll() is not None:
                    break

                # Append the output to stdout
                if output:
                    # print(output.strip())
                    stdout += output.strip() + "\n"
                    stderr += errorp.strip() + "\n"
        except KeyboardInterrupt:
            # print("KeyboardInterrupt")
            process.terminate()

        # Capture the remaining output and error (if any)
        stdout_remainder, stderr_remainder = process.communicate()
        stdout += stdout_remainder
        stderr += stderr_remainder

        return stdout, stderr


class OCR:
    commands = {"pdf2jpg": "", "jpg2pdf": ""}

    def __init__(self, source_path, ocr_lang):
        self.source_path = source_path
        self.directory, self.filename = os.path.split(source_path)
        self.ocr_output_path = "./temp/" + self.filename.split(".")[0] + ".pdf"
        self.ocr_lang = ocr_lang

        self.commands["pdf2jpg"] = (
            "pdftoppm -jpeg "
            + source_path
            + " temp/"
            + self.filename.split(".")[0]
            + "/"
        )
        self.commands["jpg2pdf"] = (
            "convert temp/"
            + self.filename.split(".")[0]
            + "/*.jpg temp/"
            + self.filename.split(".")[0]
            + "_tmp.pdf"
        )

        generate_dir = "temp/" + self.filename.split(".")[0] + "/"
        if os.path.exists(generate_dir):
            # print(f"The directory '{generate_dir}' exists, removing previous work.")
            shutil.rmtree(generate_dir)
            os.mkdir(generate_dir)
        else:
            # print(f"The directory '{generate_dir}' does not exist, creating the directory.")
            os.mkdir(generate_dir)

    def extract(self):
        try:
            # print(self.commands["pdf2jpg"])
            stdout, stderr = self.exec(self.commands["pdf2jpg"])
            # print("stdout: \n", stdout)
            # print("stderr: \n", stderr)

            # print(self.commands["jpg2pdf"])
            stdout, stderr = self.exec(self.commands["jpg2pdf"])
            # print("stdout: \n", stdout)
            # print("stderr: \n", stderr)

            ocrmypdf.configure_logging(-1)  # comment out to enable logging
            ocrmypdf.ocr(
                "./temp/" + self.filename.split(".")[0] + "_tmp.pdf",
                "./temp/" + self.filename.split(".")[0] + ".pdf",
                output_type="pdf",
                language=self.ocr_lang,
                #  force_ocr=True
            )

            shutil.rmtree("temp/" + self.filename.split(".")[0] + "/")
            os.remove("temp/" + self.filename.split(".")[0] + "_tmp.pdf")

        except KeyboardInterrupt:
            print("KeyboardInterrupt")
            exit(0)
        except Exception as e:
            print(colored(e, "red"))
            exit(0)

    def exec(self, command):
        # Execute the command and capture the output
        process = subprocess.Popen(
            command,
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )

        stdout = ""
        stderr = ""

        try:
            # Continuously read the output
            while True:
                # Read the output
                output = process.stdout.readline()
                # errorp = process.stderr.readline()

                # Check if the output is empty and the process has finished
                if process.poll() is not None:
                    break

                # Append the output to stdout
                if output:
                    # print(output.strip())
                    stdout += output.strip() + "\n"
                    # stderr += errorp.strip() + '\n'
        except KeyboardInterrupt:
            print("KeyboardInterrupt")
            process.terminate()

        # Capture the remaining output and error (if any)
        stdout_remainder, stderr_remainder = process.communicate()
        stdout += stdout_remainder
        stderr += stderr_remainder

        return stdout, stderr


class PDF:
    embeddingStore = {
        "embeddings": [],
    }
    minimumComparisonCount = 5

    def generateEmbeddingsFromFile(self, file_path):

        with open(file_path, "r", encoding="utf-8") as file:
            rawText = file.read()

        paras = []
        rawParas = rawText.split("\n\n")

        for rawPara in rawParas:
            rawPara = rawPara.strip().replace("\n", " ").replace("\r", "")
            if (
                rawPara
                and rawPara[-1] != "?"
                and len(rawPara.split()) >= self.minimumComparisonCount
            ):
                paras.append(rawPara)

        countParas = len(paras)

        # get current timestamp
        current_timestamp = int(time.time())

        response = openai.Embedding.create(input=paras, model="text-embedding-ada-002")

        # Check if data received correctly
        if len(response["data"]) >= countParas:
            for i in range(countParas):
                # Adding each embedded para to embeddingStore
                self.embeddingStore["embeddings"].append(
                    {
                        "raw": paras[i],
                        "embedding": response["data"][i]["embedding"],
                        "created": current_timestamp,
                    }
                )

        export_file_name = (
            "./temp/" + file_path.split("/")[-1].split(".")[0] + "_embeddings.json"
        )

        with open(export_file_name, "w") as file:
            json.dump(self.embeddingStore, file, indent=2)

        """
    {
      "embeddings": [
        {
          "raw": "..."
          "embedding": [0.1, 0.2, ...]
          "created": 1234567890
        },
        ...
      ]
    }
    """

    def loadEmbeddingsFromFile(self, file_path):
        with open(file_path, "r") as f:
            embeddings = json.load(f)

        self.embeddingStore = embeddings
        # print(json.dumps(self.embeddingStore, indent=2))

    def generateEmbeddingsFromText(self, text: str):
        response = openai.Embedding.create(input=text, model="text-embedding-ada-002")

        if response["data"][0]["embedding"] is None:
            print("Error: Embeddings not generated for `" + text + "`")
            return None

        # get current timestamp
        current_timestamp = int(time.time())

        embeddings = {
            "raw": text,
            "embeddings": response["data"][0]["embedding"],
            "created": current_timestamp,
        }

        # print(f"# of prompt tokens: {response['usage']['prompt_tokens']}")
        # print(f"# of total tokens: {response['usage']['total_tokens']}")

        return embeddings

    def createPrompt(self, question, paragraph):
        # Prompt taken from https://github.com/tsensei/QueryGPT
        return (
            "Answer the following question, also use your own knowledge when necessary:\n\n"
            "Context:\n"
            + "\n\n".join(paragraph)
            + "\n\nQuestion:\n"
            + question
            + "?"
            + "\n\nAnswer:"
        )

    def compareEmbeddings(self, embedding1, embedding2):
        # naive comparing of embeddings from https://github.com/tsensei/QueryGPT
        # @TODO:
        # - use KNN

        length = min(len(embedding1), len(embedding2))
        dotprod = 0

        for i in range(length):
            dotprod += embedding1[i] * embedding2[i]

        return dotprod

    def findClosestChunks(self, questionEmbedding):
        items = []

        for embedding in self.embeddingStore["embeddings"]:
            paragraph = embedding["raw"]

            currentEmbedding = embedding["embedding"]

            items.append(
                {
                    "paragraph": paragraph,
                    "score": self.compareEmbeddings(
                        questionEmbedding, currentEmbedding
                    ),
                }
            )

        items.sort(key=lambda x: x["score"], reverse=True)
        return [item["paragraph"] for item in items[: self.minimumComparisonCount]]

    def generateAnswers(self, prompt, closestParagraphs):
        try:
            completion = openai.ChatCompletion.create(
                # model="gpt-3.5-turbo",
                # model="gpt-4",
                model="gpt-5-mini",
                messages=[
                    {
                        "role": "user",
                        "content": self.createPrompt(prompt, closestParagraphs),
                    },
                ],
                # max_tokens=2000,
                temperature=0,
            )

            if len(completion["choices"]) == 0:
                print("No answers generated.")
                return None

            return True, completion["choices"][0]["message"]["content"]

        except Exception as e:
            # print(colored(e, 'red'))
            return False, e


def clearTerminal():
    # Clear command for Windows
    if os.name == "nt":
        _ = os.system("cls")
    # Clear command for Unix-like systems (Linux, macOS)
    else:
        _ = os.system("clear")



import json
from datetime import datetime
import pyperclip
import time

from termcolor import colored, cprint
from rich.console import Console
from rich import print as rprint
from rich.panel import Panel
from rich.live import Live
from rich.markdown import Markdown
from prompt_toolkit import prompt as prompt_input
from prompt_toolkit.formatted_text import ANSI

import openai
from argparse import ArgumentParser

from .utils import clearTerminal, Chat, bcolors


def main():
    #      _
    #     / \   _ __ __ _ ___
    #    / _ \ | '__/ _` / __|
    #   / ___ \| | | (_| \__ \
    #  /_/   \_\_|  \__, |___/
    #               |___/
    #
    # Initialize the argument parser
    parser = ArgumentParser()
    parser.add_argument(
        "-c",
        "--load-character",
        dest="character_file",
        help="Specify a character file location.",
    )
    parser.add_argument(
        "--stream",
        dest="steaming_mode",
        default=False,
        action="store_true",
        help="Enable streaming mode.",
    )
    parser.add_argument(
        "-e",
        "--load-engine",
        dest="engine_type",
        default="gpt-5-mini",
        help="Specify an engine type, default is `gpt-5-mini`.",
    )
    parser.add_argument(
        "--tts",
        dest="enable_tts",
        default=False,
        action="store_true",
        help="Enable text-to-speech.",
    )
    parser.add_argument(
        "-q",
        "--question",
        dest="direct_question",
        default=False,
        action="store_true",
        help="Ask a question to the chatbot and get an answer directly.",
    )
    args = parser.parse_args()

    if args.direct_question is False:
        # Clear the terminal
        clearTerminal()

    # Set up the console
    console = Console()

    #    ____ _           _
    #   / ___| |__   __ _| |_
    #  | |   | '_ \ / _` | __|
    #  | |___| | | | (_| | |_
    #   \____|_| |_|\__,_|\__|
    #
    # Initialize the chat
    chat = Chat()

    if args.character_file is not None:
        chat.loadCharacters(args.character_file)
    character = chat.character

    messages = chat.getInitialMessage()
    chat_datetime = datetime.now().strftime("%Y%m%d-%H%M%S")

    if args.direct_question is True:
        try:
            user_input = input(colored("Enter your question: ", "yellow"))
            print(bcolors.ENDC, end="")

            messages.append({"role": "user", "content": user_input})

            current_msg = ""
            for chunk in openai.ChatCompletion.create(
                model="gpt-5-mini",
                # model="gpt-4",
                messages=messages,
                # temperature=character["temperature"],
                stream=True,
            ):
                content = chunk["choices"][0].get("delta", {}).get("content")
                if content is not None:
                    current_msg += content
                    cprint(content, "light_green", end="", flush=True)

            print()
            exit(0)
        except KeyboardInterrupt:
            # Exit the program if the user presses Ctrl+C
            chat.showGoodbyeMessage()
            exit(0)
        except Exception as e:
            # Print the exception and exit the program
            print(colored(e, "red"))
            exit(0)

    chat.welcome(engine_type=args.engine_type)

    last_interrupt_time = None
    INTERRUPT_TIMEOUT = 1  # seconds

    try:
        while True:
            try:
                # Prompt for user input
                user_input = prompt_input(
                    ANSI(colored("Enter your message: ", "yellow"))
                )
                last_interrupt_time = None  # Reset on successful input
                print(bcolors.ENDC, end="")

                # Exit the program if the user types `/exit``
                if (
                    user_input == "/exit"
                    or user_input == "/quit"
                    or user_input == "/bye"
                ):
                    chat.showGoodbyeMessage()
                    exit(0)

                # Append user input to the list of messages to the previous messages
                messages.append({"role": "user", "content": user_input})

                if args.steaming_mode:
                    current_msg = ""
                    for chunk in openai.ChatCompletion.create(
                        model="gpt-3.5-turbo",
                        # model="gpt-4",
                        messages=messages,
                        temperature=character["temperature"],
                        stream=True,
                    ):
                        content = chunk["choices"][0].get("delta", {}).get("content")
                        if content is not None:
                            current_msg += content
                            cprint(content, "light_green", end="", flush=True)

                    # Append assistant response to the list of messages
                    messages.append({"role": "assistant", "content": current_msg})
                    print()
                else:
                    status = console.status(
                        f"Waiting for {character['name']}...", spinner="bouncingBar"
                    )
                    with Live(
                        Panel(status, border_style="green bold", style="green"),
                        refresh_per_second=4,
                        transient=True,
                    ):
                        try:
                            request_payload = {
                                "model": args.engine_type,
                                "messages": messages,
                                "timeout": 30,
                            }
                            if not args.engine_type.startswith("gpt-5"):
                                request_payload["temperature"] = character[
                                    "temperature"
                                ]
                            result = openai.ChatCompletion.create(**request_payload)
                        except Exception as e:
                            rprint(
                                Panel(
                                    str(e),
                                    title="System",
                                    border_style="red bold",
                                    style="red",
                                )
                            )
                            continue

                    status.stop()

                    # Copy the response to the clipboard
                    pyperclip.copy(result.choices[0].message.content)

                    response = result.choices[0].message.content
                    rprint(
                        Panel(
                            Markdown(response),
                            title=character["name"],
                            highlight=True,
                            border_style="green bold",
                            style="green",
                        )
                    )

                    rprint(
                        Panel(
                            "Copied to clipboard",
                            title="SYSTEM",
                            border_style="bright_black",
                            style="bright_black",
                        )
                    )

                    if args.enable_tts:
                        chat.say(result.choices[0].message.content, character["voice"])

                    messages.append(
                        {
                            "role": "assistant",
                            "content": result.choices[0].message.content,
                        }
                    )

                    with open(chat.getHistoryFilePath(chat_datetime), "w") as f:
                        f.writelines(json.dumps(messages, indent=2))

            except KeyboardInterrupt:
                current_time = time.time()

                # If user pressed Ctrl+C again within the timeout, exit
                if (
                    last_interrupt_time
                    and (current_time - last_interrupt_time) < INTERRUPT_TIMEOUT
                ):
                    print()
                    chat.showGoodbyeMessage()
                    exit(0)

                # First Ctrl+C - clear input and show message
                last_interrupt_time = current_time
                print(colored("\nPress Ctrl+C again to exit", "yellow"))
                continue

    except KeyboardInterrupt:
        # Exit the program if the user presses Ctrl+C
        chat.showGoodbyeMessage()
        exit(0)
    except Exception as e:
        # Print the exception and exit the program
        print(colored(e, "red"))
        exit(0)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""
CLI entry point for termq
"""

import sys
from argparse import ArgumentParser


def main():
    """Main CLI entry point that routes to chat.py or pdf.py"""
    # Parse initial args to check for --pdf flag
    parser = ArgumentParser(add_help=False)
    parser.add_argument(
        "--pdf", dest="use_pdf", action="store_true", help="Use PDF mode"
    )

    # Parse known args to check for pdf flag
    args, remaining = parser.parse_known_args()

    if args.use_pdf:
        # Import and run pdf mode
        from termq.pdf_mode import main as pdf_main

        # Re-inject remaining args for pdf_mode to parse
        sys.argv = [sys.argv[0]] + remaining
        pdf_main()
    else:
        # Import and run chat mode
        from termq.chat_mode import main as chat_main

        # Re-inject all args except the script name for chat_mode to parse
        sys.argv = [sys.argv[0]] + sys.argv[1:]
        chat_main()


if __name__ == "__main__":
    main()



import fitz
import os
from termcolor import colored
from rich.console import Console

# @TODO:
# - Save the chat history to a JSON file like what `chat.py` does

from .utils import PDF
from .utils import OCR
from .utils import clearTerminal
from argparse import ArgumentParser

# Clear the terminal
clearTerminal()

#      _
#     / \   _ __ __ _ ___
#    / _ \ | '__/ _` / __|
#   / ___ \| | | (_| \__ \
#  /_/   \_\_|  \__, |___/
#               |___/
# Initialize the argument parser
parser = ArgumentParser()
parser.add_argument(
    "-f",
    "--file",
    dest="file_path",
    required=True,  # required argument
    help="Specify a PDF file location.",
)
parser.add_argument(
    "--ocr",
    dest="need_ocr",
    action="store_true",
    default=False,
    help="Specify whether to OCR the PDF file.",
)
parser.add_argument(
    "--ocr-lang",
    dest="ocr_lang",
    default="eng",
    help="Specify the language for OCR. Default: eng",
)
args = parser.parse_args()

check_api_key = os.getenv("OPENAI_API_KEY")
if check_api_key is None:
    print(colored("OPENAI_API_KEY is not set as a system environment variable", "red"))
    exit(0)
else:
    print("[SYSTEM]", colored("OPENAI_API_KEY found.", "green"))
    print("[SYSTEM]", colored("<Press Ctrl+C to exit>", "green"))

# Initialize the console for displaying spinner status
console = Console()

# check if file exists
if not os.path.isfile(args.file_path):
    print(
        "[SYSTEM]",
        colored(
            f"The file path you specified does not exist: `{args.file_path}`", "red"
        ),
    )
    exit(0)

# check if temp directory exists
# all generated files (extracted content, embeddings) will be stored in this directory
if not os.path.isdir("temp"):
    print("Directory does not exist.")
    os.mkdir("temp")

# Initialize the PDF class
pdfUtils = PDF()

#    ____ _               _      _____           _              _     _ _
#   / ___| |__   ___  ___| | __ | ____|_ __ ___ | |__   ___  __| | __| (_)_ __   __ _ ___
#  | |   | '_ \ / _ \/ __| |/ / |  _| | '_ ` _ \| '_ \ / _ \/ _` |/ _` | | '_ \ / _` / __|
#  | |___| | | |  __/ (__|   <  | |___| | | | | | |_) |  __/ (_| | (_| | | | | | (_| \__ \
#   \____|_| |_|\___|\___|_|\_\ |_____|_| |_| |_|_.__/ \___|\__,_|\__,_|_|_| |_|\__, |___/
#                                                                               |___/
# check if the extracted txt file exists

directory, filename = os.path.split(args.file_path)

need_generate_embeddings = False
if not os.path.isfile(
    "./temp/" + filename.split(".")[0] + "_extracted_embeddings.json"
):
    need_generate_embeddings = True

#   ____  ____  _____    ___   ____ ____
#  |  _ \|  _ \|  ___|  / _ \ / ___|  _ \
#  | |_) | | | | |_    | | | | |   | |_) |
#  |  __/| |_| |  _|   | |_| | |___|  _ <
#  |_|   |____/|_|      \___/ \____|_| \_\
# (Optional)

if args.need_ocr and need_generate_embeddings is True:
    print("[SYSTEM]", colored("Starting to OCR the file.", "green"))
    with console.status(colored("OCR-ing the PDF file...", "green")) as status:
        directory, filename = os.path.split(args.file_path)
        ocr = OCR(args.file_path, args.ocr_lang)
        ocr.extract()

        ocr_output_path = "./temp/" + filename.split(".")[0] + ".pdf"
        args.file_path = ocr_output_path

#   ____  ____  _____   _          _____         _
#  |  _ \|  _ \|  ___| | |_ ___   |_   _|____  _| |_
#  | |_) | | | | |_    | __/ _ \    | |/ _ \ \/ / __|
#  |  __/| |_| |  _|   | || (_) |   | |  __/>  <| |_
#  |_|   |____/|_|      \__\___/    |_|\___/_/\_\\__|

if need_generate_embeddings is True:
    print("[SYSTEM]", colored("Extracting text from the PDF file.", "green"))
    # Setting up the PDF extraction
    doc = fitz.open(args.file_path)
    out = open(
        "./temp/" + filename.split(".")[0] + "_extracted.txt", "wb"
    )  # create a text output

    for page in doc:  # iterate the document pages
        text = page.get_text().encode("utf8")  # get plain text (is in UTF-8)
        out.write(text)  # write text of page
        out.write(bytes((12,)))  # write page delimiter (form feed 0x0C)
        out.write("\n\n\n".encode("utf8"))

    # close the writing
    out.close()

if need_generate_embeddings is True:
    print(
        "[SYSTEM]",
        colored(
            "Embeddings not found. Generating embeddings from extracted text.", "green"
        ),
    )
    pdfUtils.generateEmbeddingsFromFile(
        "./temp/" + filename.split(".")[0] + "_extracted.txt"
    )
else:
    print(
        "[SYSTEM]",
        colored("Embeddings found. Loading the existing embeddings.", "green"),
    )
    pdfUtils.loadEmbeddingsFromFile(
        "./temp/" + filename.split(".")[0] + "_extracted_embeddings.json"
    )


def main():
    try:
        while True:
            question = input(colored("Enter your question: ", "yellow"))
            question_embeddings = pdfUtils.generateEmbeddingsFromText(question)[
                "embeddings"
            ]

            generate_success = False
            while generate_success is False:
                # By reducing the length of the context tokens,
                # this loop will keep running until the GPT-3 API returns a valid response
                try:
                    closestParagraphs = pdfUtils.findClosestChunks(question_embeddings)

                    # Display indicator that the program is waiting for GPT-3
                    with console.status(
                        colored("Waiting for GPT...", "green")
                    ) as status:
                        # Retrieve the response from the API
                        generate_success, result = pdfUtils.generateAnswers(
                            question, closestParagraphs
                        )

                    if generate_success is True:
                        print(colored(result, "green"))
                    else:
                        # print(colored(result, 'red'))
                        pdfUtils.minimumComparisonCount -= 1
                except Exception as e:
                    pass

    except KeyboardInterrupt:
        # Exit the program if the user presses Ctrl+C
        print("\n[SYSTEM]", colored("Bye!", "green"))
        exit(0)
    except Exception as e:
        # Print the exception and exit the program
        print(colored(e, "red"))
        exit(0)


if __name__ == "__main__":
    main()
